#  Copyright (c) 2019 2021 by Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Isolate Python 2.6 and 2.7 version-specific semantic actions here.
"""

from uncompyle6.semantics.consts import TABLE_DIRECT

def customize_for_version26_27(self, version):

    ########################################
    # Python 2.6+
    #    except <condition> as <var>
    # vs. older:
    #    except <condition> , <var>
    #
    # For 2.6 we use the older syntax which
    # matches how we parse this in bytecode
    ########################################
    if version > (2, 6):
        TABLE_DIRECT.update({
            "except_cond2":	( "%|except %c as %c:\n", 1, 5 ),
            # When a generator is a single parameter of a function,
            # it doesn't need the surrounding parenethesis.
            "call_generator": ('%c%P', 0, (1, -1, ', ', 100)),
        })
    else:
        TABLE_DIRECT.update({
            'testtrue_then': ( 'not %p', (0, 22) ),
        })

    # FIXME: this should be a transformation
    def n_call(node):
        mapping = self._get_mapping(node)
        key = node
        for i in mapping[1:]:
            key = key[i]
            pass
        if key.kind == 'CALL_FUNCTION_1':
            # A function with one argument. If this is a generator,
            # no parenthesis is needed.
            args_node = node[-2]
            if args_node == 'expr':
                n = args_node[0]
                if n == 'generator_exp':
                    node.kind = 'call_generator'
                    pass
                pass

        self.default(node)
    self.n_call = n_call

    def n_import_from(node):
        if node[0].pattr > 0:
            node[2].pattr = ("." * node[0].pattr) + node[2].pattr
        self.default(node)
    self.n_import_from = n_import_from



# from 3.7 inspect.py
# Bug was "if not predicate or" inside "for".
# Jump optimization turns a POP_JUMP_IF_TRUE into
# a POP_JUMP_IF_FALSE and this has to be
# dealt with at the "if" (or actually "testfalse") level.

# RUNNABLE!
def getmembers(names, object, predicate):
    for key in names:
        if not predicate or object:
            object = 2
        object += 1
    return object

assert getmembers([1], 0, False) == 3
assert getmembers([1], 1, True) == 3
assert getmembers([1], 0, True) == 1
assert getmembers([1], 1, False) == 3
assert getmembers([], 1, False) == 1
assert getmembers([], 2, True) == 2

def _shadowed_dict(klass, a, b, c):
    for entry in klass:
        if not (a and b):
            c = 1
    return c

assert _shadowed_dict([1], True, True, 3) == 3
assert _shadowed_dict([1], True, False, 3) == 1
assert _shadowed_dict([1], False, True, 3) == 1
assert _shadowed_dict([1], False, False, 3) == 1
assert _shadowed_dict([], False, False, 3) == 3

# Bug: the double "and" comes out as if .. if not and
def _shadowed_dict2(klass, a, b, c, d):
    for entry in klass:
        if not (a and b and c):
            d = 1
    return d

# Not yet --
# assert _shadowed_dict2([1], False, False, False, 3) == 1
# assert _shadowed_dict2([1], True, True, True, 3) == 3



#  Copyright (c) 2018-2021 by Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Isolate Python 3 version-specific semantic actions here.
"""

from uncompyle6.semantics.consts import TABLE_DIRECT

from xdis import co_flags_is_async, iscode
from uncompyle6.scanner import Code
from uncompyle6.semantics.helper import (
    find_code_node,
    gen_function_parens_adjust,
)

from uncompyle6.semantics.make_function3 import make_function3_annotate
from uncompyle6.semantics.customize35 import customize_for_version35
from uncompyle6.semantics.customize36 import customize_for_version36
from uncompyle6.semantics.customize37 import customize_for_version37
from uncompyle6.semantics.customize38 import customize_for_version38


def customize_for_version3(self, version):
    TABLE_DIRECT.update(
        {
            "comp_for": (" for %c in %c", (2, "store"), (0, "expr")),
            "if_exp_not": (
                "%c if not %c else %c",
                (2, "expr"),
                (0, "expr"),
                (4, "expr"),
            ),
            "except_cond2": ("%|except %c as %c:\n", (1, "expr"), (5, "store")),
            "function_def_annotate": ("\n\n%|def %c%c\n", -1, 0),
            # When a generator is a single parameter of a function,
            # it doesn't need the surrounding parenethesis.
            "call_generator": ("%c%P", 0, (1, -1, ", ", 100)),
            "importmultiple": ("%|import %c%c\n", 2, 3),
            "import_cont": (", %c", 2),
            "kwarg": ("%[0]{attr}=%c", 1),
            "raise_stmt2": ("%|raise %c from %c\n", 0, 1),
            "tf_tryelsestmtl3": ( '%c%-%c%|else:\n%+%c', 1, 3, 5 ),
            "store_locals": ("%|# inspect.currentframe().f_locals = __locals__\n",),
            "with": ("%|with %c:\n%+%c%-", 0, 3),
            "withasstmt": ("%|with %c as (%c):\n%+%c%-", 0, 2, 3),
        }
    )

    assert version >= (3, 0)

    # In 2.5+ and 3.0+ "except" handlers and the "finally" can appear in one
    # "try" statement. So the below has the effect of combining the
    # "tryfinally" with statement with the "try_except" statement.
    # FIXME: something doesn't smell right, since the semantics
    # are different. See test_fileio.py for an example that shows this.
    def tryfinallystmt(node):
        suite_stmts = node[1][0]
        if len(suite_stmts) == 1 and suite_stmts[0] == 'stmt':
            stmt = suite_stmts[0]
            try_something = stmt[0]
            if try_something == "try_except":
                try_something.kind = "tf_try_except"
            if try_something.kind.startswith("tryelsestmt"):
                if try_something == "tryelsestmtl3":
                    try_something.kind = 'tf_tryelsestmtl3'
                else:
                    try_something.kind = 'tf_tryelsestmt'
        self.default(node)
    self.n_tryfinallystmt = tryfinallystmt

    def n_classdef3(node):
        """Handle "classdef" nonterminal for 3.0 >= version 3.0 < 3.6
        """

        assert (3, 0) <= self.version < (3, 6)

        # class definition ('class X(A,B,C):')
        cclass = self.currentclass

        # Pick out various needed bits of information
        # * class_name - the name of the class
        # * subclass_info - the parameters to the class  e.g.
        #      class Foo(bar, baz)
        #               ----------
        # * subclass_code - the code for the subclass body
        subclass_info = None
        if node == "classdefdeco2":
            if self.version < (3, 4):
                class_name = node[2][0].attr
            else:
                class_name = node[1][2].attr
            build_class = node
        else:
            build_class = node[0]
            class_name = node[1][0].attr
            build_class = node[0]

        assert "mkfunc" == build_class[1]
        mkfunc = build_class[1]
        if mkfunc[0] in ("kwargs", "no_kwargs"):
            if (3, 0) <= self.version < (3, 3):
                for n in mkfunc:
                    if hasattr(n, "attr") and iscode(n.attr):
                        subclass_code = n.attr
                        break
                    elif n == "expr":
                        subclass_code = n[0].attr
                    pass
                pass
            else:
                for n in mkfunc:
                    if hasattr(n, "attr") and iscode(n.attr):
                        subclass_code = n.attr
                        break
                    pass
                pass
            if node == "classdefdeco2":
                subclass_info = node
            else:
                subclass_info = node[0]
        elif build_class[1][0] == "load_closure":
            # Python 3 with closures not functions
            load_closure = build_class[1]
            if hasattr(load_closure[-3], "attr"):
                # Python 3.3 classes with closures work like this.
                # Note have to test before 3.2 case because
                # index -2 also has an attr.
                subclass_code = find_code_node(load_closure, -3).attr
            elif hasattr(load_closure[-2], "attr"):
                # Python 3.2 works like this
                subclass_code = find_code_node(load_closure, -2).attr
            else:
                raise "Internal Error n_classdef: cannot find class body"
            if hasattr(build_class[3], "__len__"):
                if not subclass_info:
                    subclass_info = build_class[3]
            elif hasattr(build_class[2], "__len__"):
                subclass_info = build_class[2]
            else:
                raise "Internal Error n_classdef: cannot superclass name"
        elif not subclass_info:
            if mkfunc[0] in ("no_kwargs", "kwargs"):
                subclass_code = mkfunc[1].attr
            else:
                subclass_code = mkfunc[0].attr
            if node == "classdefdeco2":
                subclass_info = node
            else:
                subclass_info = node[0]

        if node == "classdefdeco2":
            self.write("\n")
        else:
            self.write("\n\n")

        self.currentclass = str(class_name)
        self.write(self.indent, "class ", self.currentclass)

        self.print_super_classes3(subclass_info)
        self.println(":")

        # class body
        self.indent_more()
        self.build_class(subclass_code)
        self.indent_less()

        self.currentclass = cclass
        if len(self.param_stack) > 1:
            self.write("\n\n")
        else:
            self.write("\n\n\n")

        self.prune()

    self.n_classdef3 = n_classdef3

    if version == 3.0:
        # In Python 3.0 there is code to move from _[dd] into
        # the iteration variable. These rules we can ignore
        # since we pick up the iteration variable some other way and
        # we definitely don't include in the source  _[dd].
        TABLE_DIRECT.update({
            "ifstmt30":	( "%|if %c:\n%+%c%-",
                          (0, "testfalse_then"),
                          (1, "_ifstmts_jump30") ),
            "ifnotstmt30": ( "%|if not %c:\n%+%c%-",
                             (0, "testtrue_then"),
                             (1, "_ifstmts_jump30") ),
            "try_except30": ( "%|try:\n%+%c%-%c\n\n",
                              (1, "suite_stmts_opt"),
                              (4, "except_handler") ),

            })

        def n_comp_iter(node):
            if node[0] == "expr":
                n = node[0][0]
                if n == "LOAD_FAST" and n.pattr[0:2] == "_[":
                    self.prune()
                    pass
                pass
            # Not this special case, proceed as normal...
            self.default(node)

        self.n_comp_iter = n_comp_iter

    elif version == 3.3:
        # FIXME: perhaps this can be folded into the 3.4+ case?
        def n_yield_from(node):
            assert node[0] == "expr"
            if node[0][0] == "get_iter":
                # Skip over yield_from.expr.get_iter which adds an
                # extra iter(). Maybe we can do in tranformation phase instead?
                template = ("yield from %c", (0, "expr"))
                self.template_engine(template, node[0][0])
            else:
                template = ("yield from %c", (0, "attribute"))
                self.template_engine(template, node[0][0][0])
            self.prune()

        self.n_yield_from = n_yield_from

    if (3, 2) <= version <= (3, 4):

        def n_call(node):

            mapping = self._get_mapping(node)
            key = node
            for i in mapping[1:]:
                key = key[i]
                pass
            if key.kind.startswith("CALL_FUNCTION_VAR_KW"):
                # We may want to fill this in...
                # But it is distinct from CALL_FUNCTION_VAR below
                pass
            elif key.kind.startswith("CALL_FUNCTION_VAR"):
                # CALL_FUNCTION_VAR's top element of the stack contains
                # the variable argument list, then comes
                # annotation args, then keyword args.
                # In the most least-top-most stack entry, but position 1
                # in node order, the positional args.
                argc = node[-1].attr
                nargs = argc & 0xFF
                kwargs = (argc >> 8) & 0xFF
                # FIXME: handle annotation args
                if kwargs != 0:
                    # kwargs == 0 is handled by the table entry
                    # Should probably handle it here though.
                    if nargs == 0:
                        template = ("%c(*%c, %C)", 0, -2, (1, kwargs + 1, ", "))
                    else:
                        template = (
                            "%c(%C, *%c, %C)",
                            0,
                            (1, nargs + 1, ", "),
                            -2,
                            (-2 - kwargs, -2, ", "),
                        )
                    self.template_engine(template, node)
                    self.prune()
            else:
                gen_function_parens_adjust(key, node)
            self.default(node)

        self.n_call = n_call
    elif version < (3, 2):

        def n_call(node):
            mapping = self._get_mapping(node)
            key = node
            for i in mapping[1:]:
                key = key[i]
                pass
            gen_function_parens_adjust(key, node)
            self.default(node)

        self.n_call = n_call

    def n_mkfunc_annotate(node):

        # Handling EXTENDED_ARG before MAKE_FUNCTION ...
        i = -1 if node[-2] == "EXTENDED_ARG" else 0

        if self.version < (3, 3):
            code = node[-2 + i]
        elif self.version >= (3, 3) or node[-2] == "kwargs":
            # LOAD_CONST code object ..
            # LOAD_CONST        'x0'  if >= 3.3
            # EXTENDED_ARG
            # MAKE_FUNCTION ..
            code = node[-3 + i]
        elif node[-3] == "expr":
            code = node[-3][0]
        else:
            # LOAD_CONST code object ..
            # MAKE_FUNCTION ..
            code = node[-3]

        self.indent_more()
        for annotate_last in range(len(node) - 1, -1, -1):
            if node[annotate_last] == "annotate_tuple":
                break

        # FIXME: the real situation is that when derived from
        # function_def_annotate we the name has been filled in.
        # But when derived from funcdefdeco it hasn't Would like a better
        # way to distinquish.
        if self.f.getvalue()[-4:] == "def ":
            self.write(code.attr.co_name)

        # FIXME: handle and pass full annotate args
        make_function3_annotate(
            self, node, is_lambda=False, code_node=code, annotate_last=annotate_last
        )

        if len(self.param_stack) > 1:
            self.write("\n\n")
        else:
            self.write("\n\n\n")
        self.indent_less()
        self.prune()  # stop recursing

    self.n_mkfunc_annotate = n_mkfunc_annotate

    TABLE_DIRECT.update(
        {
            "tryelsestmtl3": (
                "%|try:\n%+%c%-%c%|else:\n%+%c%-",
                (1, "suite_stmts_opt"),
                3, # "except_handler_else" or "except_handler"
                (5, "else_suitel"),
            ),
            "LOAD_CLASSDEREF": ("%{pattr}",),
        }
    )
    if version >= (3, 4):
        #######################
        # Python 3.4+ Changes #
        #######################
        TABLE_DIRECT.update(
            {
                "LOAD_CLASSDEREF": ("%{pattr}",),
                "yield_from": ("yield from %c", (0, "expr")),
            }
        )
        if version >= (3, 5):
            customize_for_version35(self, version)
            if version >= (3, 6):
                customize_for_version36(self, version)
                if version >= (3, 7):
                    customize_for_version37(self, version)
                    if version >= (3, 8):
                        customize_for_version38(self, version)
                        pass  # version >= 3.8
                    pass  # 3.7
                pass  # 3.6
            pass  # 3.5
        pass  # 3.4
    return



# Python 1.4 cgi.py
# Bug was in "continue" detection.
# 1.4 doesn't have lnotab and our CONTINUE detection is off.
def parse_multipart(params, pdict):
    while params:
	if params.has_key('name'):
	   params = None
	else:
	    continue

    return None



# Bug from 3.3 configparser.py
# Python 3.3 has in bytecode positional args after keyoard args
# Python 3.4+ has positional args before keyword args
def __init__(self, defaults=None, dict_type=_default_dict,
             allow_no_value=False, *, delimiters=('=', ':'),
             comment_prefixes=('#', ';'), inline_comment_prefixes=None,
             strict=True, empty_lines_in_values=True,
             default_section=DEFAULTSECT,
             interpolation=_UNSET):
    pass

# From 3.5 sqlalchemy/orm/__init__.py
# Python 3.5 changes the stack position of where * args are (furthest down the stack)
# Python 3.6+ replaces CALL_FUNCTION_VAR_KW with CALL_FUNCTION_EX
def deferred(*columns, **kw):
    return ColumnProperty(deferred=True, *columns, **kw)


# From sqlalchemy/sql/selectable.py
class GenerativeSelect():
    def __init__(self,
                 ClauseList,
                 util,
                 order_by=None):
        self._order_by_clause = ClauseList(
            *util.to_list(order_by),
            _literal_as_text=5)



"""
test_loops2.py -- source test pattern for loops (CONTINUE_LOOP)

This source is part of the decompyle test suite.

decompyle is a Python byte-code decompiler
See http://www.goebel-consult.de/decompyle/ for download and
for further information
"""

# This is a separate test pattern, since 'continue' within 'try'
# was not allowed till Python 2.1

for term in args:
    try:
        print
        continue
        print
    except:
        pass



# From 3.7.6 test_buffer.py

# RUNNABLE!
def foo(n):
    zero_stride = True if n >= 95 and n & 1 else False
    return zero_stride

assert foo(95)
assert not foo(94)
assert not foo(96)

# from test_buffer.py
# Bug was handling "or" inside a conditional
def rslice(a, b):
    minlen = 0 if a or b else 1
    return minlen

assert rslice(False, False) == 1
assert rslice(False, True) == 0
assert rslice(True, False) == 0
assert rslice(True, True) == 0



#  Copyright (c) 2018-2019, 2021-2022 by Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Isolate Python version-specific semantic actions here.
"""

from uncompyle6.parsers.treenode import SyntaxTree
from uncompyle6.semantics.consts import (
    INDENT_PER_LEVEL,
    NO_PARENTHESIS_EVER,
    PRECEDENCE,
    TABLE_R,
    TABLE_DIRECT,
)
from uncompyle6.semantics.helper import flatten_list
from uncompyle6.scanners.tok import Token


def customize_for_version(self, is_pypy, version):
    if is_pypy:
        ########################
        # PyPy changes
        #######################
        # fmt: off
        TABLE_DIRECT.update({

            "assert":           ("%|assert %c\n", 0),
            # This can happen as a result of an if transformation
            "assert2":          ("%|assert %c, %c\n", 0, 3),
            "assert_pypy":	    ( "%|assert %c\n" ,     (1, "assert_expr") ),
            # This is as a result of an if transformation
            'assert0_pypy':	    ( "%|assert %c\n" ,     0),

            'assert_not_pypy':	( "%|assert not %c\n" , (1, "assert_exp") ),
            "assert2_not_pypy": (
                    "%|assert not %c, %c\n",
                    (1, "assert_exp"),
                    (4, "expr"),
                ),

            "try_except_pypy":	   ( "%|try:\n%+%c%-%c\n\n", 1, 2 ),
            "tryfinallystmt_pypy": ( "%|try:\n%+%c%-%|finally:\n%+%c%-\n\n", 1, 3 ),
            "assign3_pypy":        ( "%|%c, %c, %c = %c, %c, %c\n", 5, 4, 3, 0, 1, 2 ),
            "assign2_pypy":        ( "%|%c, %c = %c, %c\n", 3, 2, 0, 1),
            })
        # fmt: on

        if version[:2] >= (3, 7):

            def n_call_kw_pypy37(node):
                self.template_engine(("%p(", (0, NO_PARENTHESIS_EVER)), node)
                assert node[-1] == "CALL_METHOD_KW"
                arg_count = node[-1].attr
                kw_names = node[-2]
                assert kw_names == "pypy_kw_keys"

                kwargs_names = kw_names[0].attr
                kwarg_count = len(kwargs_names)
                pos_argc = arg_count - kwarg_count

                flat_elems = flatten_list(node[1:-2])
                n = len(flat_elems)
                assert n == arg_count, "n: %s, arg_count: %s\n%s" % (
                    n,
                    arg_count,
                    node,
                )

                sep = ""

                for i in range(pos_argc):
                    elem = flat_elems[i]
                    line_number = self.line_number
                    value = self.traverse(elem)
                    if line_number != self.line_number:
                        sep += "\n" + self.indent + INDENT_PER_LEVEL[:-1]
                        pass
                    self.write(f"{sep}{value}")
                    sep = ", "

                assert n >= len(kwargs_names)
                j = pos_argc
                for i in range(kwarg_count):
                    elem = flat_elems[j]
                    j += 1
                    assert elem == "expr"
                    line_number = self.line_number
                    value = self.traverse(elem)
                    if line_number != self.line_number:
                        sep += "\n" + self.indent + INDENT_PER_LEVEL[:-1]
                        pass
                    self.write(f"{sep}{kwargs_names[i]}={value}")
                    sep = ", "
                    pass

                self.write(")")
                self.prune()

            self.n_call_kw_pypy37 = n_call_kw_pypy37

    else:
        ########################
        # Without PyPy
        #######################
        TABLE_DIRECT.update(
            {
                # "assert" and "assert_expr" are added via transform rules.
                "assert": ("%|assert %c\n", 0),
                "assert2": ("%|assert %c, %c\n", 0, 3),
                # Created only via transformation
                "assertnot": ("%|assert not %p\n", (0, PRECEDENCE["unary_not"])),
                "assert2not": (
                    "%|assert not %p, %c\n",
                    (0, PRECEDENCE["unary_not"]),
                    3,
                ),
                "assign2": ("%|%c, %c = %c, %c\n", 3, 4, 0, 1),
                "assign3": ("%|%c, %c, %c = %c, %c, %c\n", 5, 6, 7, 0, 1, 2),
                "try_except": ("%|try:\n%+%c%-%c\n\n", 1, 3),
            }
        )
    if version >= (3, 0):
        if version >= (3, 2):
            TABLE_DIRECT.update(
                {"del_deref_stmt": ("%|del %c\n", 0), "DELETE_DEREF": ("%{pattr}", 0)}
            )
        from uncompyle6.semantics.customize3 import customize_for_version3

        customize_for_version3(self, version)
    else:  # < 3.0
        TABLE_DIRECT.update(
            {"except_cond3": ("%|except %c, %c:\n", (1, "expr"), (-2, "store"))}
        )
        if version <= (2, 6):
            TABLE_DIRECT["testtrue_then"] = TABLE_DIRECT["testtrue"]

        if (2, 4) <= version <= (2, 6):
            TABLE_DIRECT.update({"comp_for": (" for %c in %c", 3, 1)})
        else:
            TABLE_DIRECT.update({"comp_for": (" for %c in %c%c", 2, 0, 3)})

        if version >= (2, 5):
            from uncompyle6.semantics.customize25 import customize_for_version25

            customize_for_version25(self, version)

            if version >= (2, 6):
                from uncompyle6.semantics.customize26_27 import (
                    customize_for_version26_27,
                )

                customize_for_version26_27(self, version)
                pass
        else:  # < 2.5
            global NAME_MODULE
            NAME_MODULE = SyntaxTree(
                "stmt",
                [
                    SyntaxTree(
                        "assign",
                        [
                            SyntaxTree(
                                "expr",
                                [
                                    Token(
                                        "LOAD_GLOBAL",
                                        pattr="__name__",
                                        offset=0,
                                        has_arg=True,
                                    )
                                ],
                            ),
                            SyntaxTree(
                                "store",
                                [
                                    Token(
                                        "STORE_NAME",
                                        pattr="__module__",
                                        offset=3,
                                        has_arg=True,
                                    )
                                ],
                            ),
                        ],
                    )
                ],
            )
            TABLE_DIRECT.update(
                {
                    "importmultiple": ("%|import %c%c\n", 2, 3),
                    "import_cont": (", %c", 2),
                    "tryfinallystmt": (
                        "%|try:\n%+%c%-%|finally:\n%+%c%-",
                        (1, "suite_stmts_opt"),
                        (5, "suite_stmts_opt"),
                    ),
                }
            )
            if version == (2, 4):

                def n_iftrue_stmt24(node):
                    self.template_engine(("%c", 0), node)
                    self.default(node)
                    self.prune()

                self.n_iftrue_stmt24 = n_iftrue_stmt24
            elif version < (1, 4):
                from uncompyle6.semantics.customize14 import customize_for_version14

                customize_for_version14(self, version)

                def n_call(node):
                    expr = node[0]
                    assert expr == "expr"
                    params = node[1]
                    if params == "tuple":
                        self.template_engine(("%p(", (0, NO_PARENTHESIS_EVER)), expr)
                        sep = ""
                        for param in params[:-1]:
                            self.write(sep)
                            self.preorder(param)
                            sep = ", "
                        self.write(")")
                    else:
                        self.template_engine(
                            (
                                "%p(%P)",
                                (0, "expr", 100),
                                (1, -1, ", ", NO_PARENTHESIS_EVER),
                            ),
                            node,
                        )
                    self.prune()

                self.n_call = n_call

            else:  # 1.0 <= version <= 2.3:
                TABLE_DIRECT.update({"if1_stmt": ("%|if 1\n%+%c%-", 5)})
                if version <= (2, 1):
                    TABLE_DIRECT.update(
                        {
                            "importmultiple": ("%c", 2),
                            # FIXME: not quite right. We have indiividual imports
                            # when there is in fact one: "import a, b, ..."
                            "imports_cont": ("%C%,", (1, 100, "\n")),
                        }
                    )
                    pass
                pass
            pass  # < 2.5

        # < 3.0 continues

        TABLE_R.update(
            {
                "STORE_SLICE+0": ("%c[:]", 0),
                "STORE_SLICE+1": ("%c[%p:]", 0, (1, -1)),
                "STORE_SLICE+2": ("%c[:%p]", 0, (1, -1)),
                "STORE_SLICE+3": ("%c[%p:%p]", 0, (1, -1), (2, -1)),
                "DELETE_SLICE+0": ("%|del %c[:]\n", 0),
                "DELETE_SLICE+1": ("%|del %c[%c:]\n", 0, 1),
                "DELETE_SLICE+2": ("%|del %c[:%c]\n", 0, 1),
                "DELETE_SLICE+3": ("%|del %c[%c:%c]\n", 0, 1, 2),
            }
        )
        TABLE_DIRECT.update({"raise_stmt2": ("%|raise %c, %c\n", 0, 1)})

        # exec as a built-in statement is only in Python 2.x
        def n_exec_stmt(node):
            """
            exec_stmt ::= expr exprlist DUP_TOP EXEC_STMT
            exec_stmt ::= expr exprlist EXEC_STMT
            """
            self.write(self.indent, "exec ")
            self.preorder(node[0])
            if not node[1][0].isNone():
                sep = " in "
                for subnode in node[1]:
                    self.write(sep)
                    sep = ", "
                    self.preorder(subnode)
            self.println()
            self.prune()  # stop recursing

        self.n_exec_smt = n_exec_stmt

        pass  # < 3.0

    return



# From python 2.5 make_decorators.py
# Bug was in not recognizing @memoize which uses grammar rules
# using nonterminals mkfuncdeco and mkfuncdeco0

# This file is RUNNABLE!
def memoize(func):
    pass

def test_memoize(self):
    @memoize
    def double(x):
        return x * 2

# Seen in 3.7 test/test_c_locale_coercion.py
# Bug was handling multiple decorators in 3.5+
# simply because we didn't carry over parser rules over from
# earlier versions.

x = 1
def decorator(func):
    def inc_x():
        global x
        x += 1
        func()
    return inc_x

@decorator
@decorator
def fn():
    return

assert x == 1
fn()
assert x == 3



# listComprehensions.py -- source test pattern for list comprehensions
#
# This simple program is part of the decompyle test suite.
#
# decompyle is a Python byte-code decompiler
# See http://www.goebel-consult.de/decompyle/ for further information

print [i*j for i in range(4)
       if i == 2
       for j in range(7)
       if (i+i % 2) == 0 ]



#!/usr/bin/env python
from xdis.version_info import PYTHON_VERSION_TRIPLE, IS_PYPY, version_tuple_to_str
from uncompyle6.scanner import get_scanner
def bug(state, slotstate):
    if state:
        if slotstate is not None:
            for key, value in slotstate.items():
                setattr(state, key, 2)

# From 2.7 disassemble
# Problem is not getting while, because
# COME_FROM not added
def bug_loop(disassemble, tb=None):
    if tb:
        try:
            tb = 5
        except AttributeError:
            raise RuntimeError
        while tb: tb = tb.tb_next
    disassemble(tb)

def test_if_in_for():
    code = bug.__code__
    scan = get_scanner(PYTHON_VERSION_TRIPLE)
    if (2, 7) <= PYTHON_VERSION_TRIPLE < (3, 1) and not IS_PYPY:
        scan.build_instructions(code)
        fjt = scan.find_jump_targets(False)

        ## FIXME: the data below is wrong.
        ## we get different results currenty as well.
        ## We need to probably fix both the code
        ## and the test below
        # assert {15: [3], 69: [66], 63: [18]} == fjt
        # assert scan.structs == \
        #   [{'start': 0, 'end': 72, 'type': 'root'},
        #    {'start': 15, 'end': 66, 'type': 'if-then'},
        #    {'start': 31, 'end': 59, 'type': 'for-loop'},
        #    {'start': 62, 'end': 63, 'type': 'for-else'}]

        code = bug_loop.__code__
        scan.build_instructions(code)
        fjt = scan.find_jump_targets(False)
        assert{64: [42], 67: [42, 42], 42: [16, 41], 19: [6]} == fjt
        assert scan.structs == [
            {'start': 0, 'end': 80, 'type': 'root'},
            {'start': 3, 'end': 64, 'type': 'if-then'},
            {'start': 6, 'end': 15, 'type': 'try'},
            {'start': 19, 'end': 38, 'type': 'except'},
            {'start': 45, 'end': 67, 'type': 'while-loop'},
            {'start': 70, 'end': 64, 'type': 'while-else'},
            # previous bug was not mistaking while-loop for if-then
            {'start': 48, 'end': 67, 'type': 'while-loop'}]

    elif (3, 2) < PYTHON_VERSION_TRIPLE <= (3, 4):
        scan.build_instructions(code)
        fjt  = scan.find_jump_targets(False)
        assert {69: [66], 63: [18]} == fjt
        assert scan.structs == \
          [{'end': 72, 'type': 'root', 'start': 0},
           {'end': 66, 'type': 'if-then', 'start': 6},
           {'end': 63, 'type': 'if-then', 'start': 18},
           {'end': 59, 'type': 'for-loop', 'start': 31},
           {'end': 63, 'type': 'for-else', 'start': 62}]
    else:
        print("FIXME: should fix for %s" % version_tuple_to_str())
        assert True
    return



# From python3.8/distutils/version.py with optimization -O2
# The bug was that the other "else" constant propagated removed.

# NOTE: this program needs to be compile with optimization
def _cmp (b, c):
    if b:
        if c:
            return 0
        else:
            return 1
    else:
        assert False, "never get here"



# Adapted from Python 2.4 bdb.py runeval()

# In Python 2.4 and before, try/finally has to be one block
# and try/except has to be in a separate block.

# In Python 2.5 and later, these can be combined into one "try" block,
# and indeed compiling this in 2.5+ will in fact combine the blocks.
# And that's okay, even if it might not be what was written.

# However for 2.4 and before make sure this _isn't_ combined into one block.
try:
    try:
        quitting = eval("1+2")
    except RuntimeError:
        pass
finally:
    quitting = 1



# From 2.6.9 abc.py

# For 2.6:
# genexpr_func ::= setup_loop_lf FOR_ITER store comp_iter JUMP_BACK come_from_pop JUMP_BACK POP_BLOCK COME_FROM

# This has been a bug in other Pythons after 2.6 were set comprehension {} is used instead of set().
abstracts = set(name
             for name, value in namespace.items()
             if getattr(value, "__isabstractmethod__", False))



#  Copyright (c) 2020 Rocky Bernstein

def except_handler(self, lhs, n, rule, ast, tokens, first, last):
    end_token = tokens[last-1]

    # print("XXX", first, last)
    # for t in range(first, last):
    #     print(tokens[t])
    # print("=" * 30)

    # FIXME: Figure out why this doesn't work on
    # bytecode-1.4/anydbm.pyc
    if self.version[:2] == (1, 4):
        return False

    # Make sure come froms all come from within "except_handler".
    if end_token != "COME_FROM":
        return False
    return end_token.attr < tokens[first].offset



from uncompyle6.parsers.reducecheck.and_check import *  # noqa
from uncompyle6.parsers.reducecheck.aug_assign import *  # noqa
from uncompyle6.parsers.reducecheck.except_handler import *  # noqa
from uncompyle6.parsers.reducecheck.except_handler_else import *  # noqa
from uncompyle6.parsers.reducecheck.ifelsestmt import *  # noqa
from uncompyle6.parsers.reducecheck.ifelsestmt2 import *  # noqa
from uncompyle6.parsers.reducecheck.iflaststmt import *  # noqa
from uncompyle6.parsers.reducecheck.ifstmt import *  # noqa
from uncompyle6.parsers.reducecheck.ifstmt2 import *  # noqa
from uncompyle6.parsers.reducecheck.ifstmts_jump import *  # noqa
from uncompyle6.parsers.reducecheck.for_block_check import *  # noqa
from uncompyle6.parsers.reducecheck.or_check import *  # noqa
from uncompyle6.parsers.reducecheck.testtrue import *  # noqa
from uncompyle6.parsers.reducecheck.tryelsestmt import *  # noqa
from uncompyle6.parsers.reducecheck.tryexcept import *  # noqa
from uncompyle6.parsers.reducecheck.tryelsestmtl3 import *  # noqa
from uncompyle6.parsers.reducecheck.while1elsestmt import *  # noqa
from uncompyle6.parsers.reducecheck.while1stmt import *  # noqa



# Self-checking test.
# String interpolation tests

# RUNNABLE!
"""This program is self-checking!"""

var1 = "x"
var2 = "y"
abc = "def"
assert (
    f"interpolate {var1} strings {var2!r} {var2!s} 'py36"
    == "interpolate x strings 'y' y 'py36"
)
assert "def0" == f"{abc}0"
assert "defdef" == f"{abc}{abc!s}"

# From 3.8 test/test_string.py
# We had the precedence of yield vs. lambda incorrect.
def fn(x):
    yield f"x:{yield (lambda i: x * i)}"


# From 3.6 functools.py
# Bug was handling format operator strings.

k, v = "1", ["2"]
x = f"{k}={v!r}"
y = f"functools.{x}({', '.join(v)})"
assert x == "1=['2']"
assert y == "functools.1=['2'](2)"

# From 3.6 http/client.py
# Bug is in handling  X
chunk = ["a", "b", "c"]
chunk2 = "d"
chunk = f"{len(chunk):X}" + chunk2
assert chunk == "3d"

chunk = b"abc"
chunk2 = "d"
chunk = f"{len(chunk):X}\r\n".encode("ascii") + chunk + b"\r\n"
assert chunk == b"3\r\nabc\r\n"

# From 3.6.8 idlelib/pyshell.py
# Bug was handling '''
import os

filename = "."
source = "foo"
source = f"__file__ = r'''{os.path.abspath(filename)}'''\n" + source + "\ndel __file__"

# Note how { and } are *not* escaped here
f = "one"
name = "two"
assert (f"{f}{'{{name}}'} {f}{'{name}'}") == "one{{name}} one{name}"

# From 3.7.3 dataclasses.py
log_rounds = 5
assert "05$" == f"{log_rounds:02d}$"


def testit(a, b, ll):
    # print(ll)
    return ll


# The call below shows the need for BUILD_STRING to count expr arguments.
# Also note that we use {{ }} to escape braces in contrast to the example
# above.
def _repr_fn(fields):
    return testit(
        "__repr__",
        ("self",),
        ['return xx + f"(' + ", ".join([f"{f}={{self.{f}!r}}" for f in fields]) + ')"'],
    )


fields = ["a", "b", "c"]
assert _repr_fn(fields) == ['return xx + f"(a={self.a!r}, b={self.b!r}, c={self.c!r})"']


#################################
# From Python 3.7 test_fstring.py

x = 5
assert f'{(lambda y:x*y)("8")!r}' == "'88888'"
assert f'{(lambda y:x*y)("8")!r:10}' == "'88888'   "
assert f'{(lambda y:x*y)("8"):10}' == "88888     "

try:
    eval("f'{lambda x:x}'")
except SyntaxError:
    pass
else:
    assert False, "f'{lambda x:x}' should be a syntax error"

(x, y, width) = ("foo", 2, 10)
assert f"x={x*y:{width}}" == "x=foofoo    "

# Why the fact that the distinction of docstring versus stmt is a
# string expression is important academic, but we will decompile an
# equivalent thing. For compatiblity with older Python we'll use "%"
# instead of a format string
def f():
    f"""Not a docstring"""  # noqa


def g():
    """Not a docstring""" f""  # noqa


assert f.__doc__ is None
assert g.__doc__ is None

import decimal

width, precision, value = (10, 4, decimal.Decimal("12.34567"))

# Make sure we don't have additional f'..' inside the format strings below.
assert f"result: {value:{width}.{precision}}" == "result:      12.35"
assert f"result: {value:{width:0}.{precision:1}}" == "result:      12.35"
assert f"{2}\t" == "2\t"

# But below we *do* need the additional f".."
assert f'{f"{0}"*3}' == "000"

# We need to make sure we have { {x:... not {{x: ...
#                               ^
# The former, {{ confuses the format strings so dictionary/set comprehensions
# don't work.
assert f"expr={ {x: y for x, y in [(1, 2), ]}}" == "expr={1: 2}"


class Line:
    def __init__(self, x, y):
        self.x = x
        self.y = y

    # From 3.7 test_typing.py
    def __str__(self):
        return f"{self.x} -> {self.y}"


line = Line(1, 2)
assert str(line) == "1 -> 2"



"""This program is self-checking!"""


class C:
    def sort(self, l, reverse, key_fn):
        # PyPy example of CALL_METHOD_KW 2
        # 2 keyword arguments and no positional arguments
        return l.sort(reverse=reverse, key=key_fn)


def lcase(s):
    return s.lower()


x = C()
l = ["xyz", "ABC"]

# An PyPy example of CALL_METHOD_KW 3
# 2 keyword arguments and one positional argument
x.sort(l, reverse=False, key_fn=lcase)
assert l == ["ABC", "xyz"]



import os.path
import pytest

from uncompyle6.code_fns import disassemble_file

def get_srcdir():
    filename = os.path.normcase(os.path.dirname(__file__))
    return os.path.realpath(filename)

src_dir = get_srcdir()
os.chdir(src_dir)


@pytest.mark.parametrize(("test_tuple"), [
    ('../test/bytecode_2.7/05_if.pyc', 'testdata/if-2.7.right',),
    ('../test/bytecode_2.7/05_ifelse.pyc', 'testdata/ifelse-2.7.right',),
])
def test_funcoutput(capfd, test_tuple):

    in_file, filename_expected = test_tuple
    disassemble_file(in_file)
    resout, reserr = capfd.readouterr()
    expected = open(filename_expected, "r").read()
    if resout != expected:
        with open(filename_expected + ".got", "w") as out:
            out.write(resout)
    assert resout == expected



#!/usr/bin/env python
# See https://github.com/rocky/python-uncompyle6/pull/15

# In Python 2.7, you should see
#   mkfuncdeco0 ::= mkfunc
#   classdefdeco2 ::= LOAD_CONST expr mkfunc CALL_FUNCTION_0 BUILD_CLASS
#   classdefdeco1 ::= expr classdefdeco1 CALL_FUNCTION_1
#   store        ::= STORE_NAME
#   classdefdeco ::= classdefdeco1 store

def author(*author_names):
    def author_func(cls):
        return cls
    return author_func

@author('Me', 'Him')
@author('You')
class MyClass(object):
    def __init__(self):
        pass

    @staticmethod
    @staticmethod
    def static_method():
        pass

x = MyClass()

# Try class without parens
class Feature:
    pass



# Adapted from 1.4 anydbm
"""This program is self-checking!"""
def scan(items):
    for i in items:
            try:
                    5 / i
            except:
                    continue
            else:
                    break
    else:
            return 2
    return i

assert scan((0, 1)) == 1
assert scan((0, 0)) == 2
assert scan((3, 2, 1)) == 3



#  Copyright (c) 2016-2017, 2021-2022 by Rocky Bernstein
"""
Python 2.4 bytecode massaging.

This massages tokenized 2.7 bytecode to make it more amenable for
grammar parsing.
"""

import uncompyle6.scanners.scanner25 as scan

# bytecode verification, verify(), uses JUMP_OPs from here
from xdis.opcodes import opcode_24

JUMP_OPS = opcode_24.JUMP_OPS

# We base this off of 2.5 instead of the other way around
# because we cleaned things up this way.
# The history is that 2.7 support is the cleanest,
# then from that we got 2.6 and so on.
class Scanner24(scan.Scanner25):
    def __init__(self, show_asm=False):
        scan.Scanner25.__init__(self, show_asm)
        # These are the only differences in initialization between
        # 2.4, 2.5 and 2.6
        self.opc = opcode_24
        self.opname = opcode_24.opname
        self.version = (2, 4)
        self.genexpr_name = "<generator expression>"
        return


if __name__ == "__main__":
    from xdis.version_info import PYTHON_VERSION_TRIPLE, version_tuple_to_str

    if PYTHON_VERSION_TRIPLE[:2] == (2, 4):
        import inspect

        co = inspect.currentframe().f_code  # type: ignore
        tokens, customize = Scanner24().ingest(co)
        for t in tokens:
            print(t.format())
        pass
    else:
        print("Need to be Python 2.4 to demo; I am version %s" % version_tuple_to_str())



# from 3.7 test_contextlib_async.py
# Bugs were not adding "async" when a function is a decorator,
# and a misaligment when using "async with ... as".

"""This program is self-checking!"""
import asyncio
from contextlib import asynccontextmanager, AbstractAsyncContextManager
import functools


def _async_test(func):
    """Decorator to turn an async function into a test case."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        coro = func(*args, **kwargs)
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        try:
            return loop.run_until_complete(coro)
        finally:
            loop.close()
            asyncio.set_event_loop(None)

    return wrapper


state = []


@asynccontextmanager
async def woohoo():
    state.append(1)
    yield 42
    state.append(999)


@_async_test
async def test_enter():
    class DefaultEnter(AbstractAsyncContextManager):
        async def __aexit__(*args):
            return
            # await super().__aexit__(*args)

    manager = DefaultEnter()
    got_manager = await manager.__aenter__()
    # print(got_manager, manager)
    assert got_manager is manager

    async with manager as context:
        async with woohoo() as x:
            x = 1
            y = 2
        assert manager is context


# From 3.7.6 test_coroutines.py
# Bug was different form of code for "async with" below
class CoroutineTest:
    def test_with_8(self):
        CNT = 0

        async def foo():
            nonlocal CNT
            async with CM():
                CNT += 1
                return


test_enter()



# Tests all the different kinds of imports
"""This program is self-checking!"""

import sys
from os import path
from os import *  # NOQA

import time as time1, os as os1

assert isinstance(os1.pathsep, str)

assert time1.time() > 0
import os.path as osp

assert osp == path
from os.path import join as jj

assert path.join("a", "b") == jj("a", "b")

if len(__file__) == 0:
    # a.b.c should force consecutive LOAD_ATTRs
    import stuff0.stuff1.stuff2.stuff3 as stuff3

sys.exit(0)



# Issue #149. Bug in Python 2.7 was handling a return stmt at the end
# of a while with so no jump back, confusing which block the
# return should be part of
def test(a):
    while True:
        if a:
            pass
        else:
            continue
        return



# From Python 3.4 calendar

# load_closure ::= LOAD_CLOSURE
# load_closure ::= load_closure LOAD_CLOSURE
# expr ::= LOAD_GLOBAL
# get_iter ::= expr GET_ITER
# genexpr ::= load_closure LOAD_GENEXPR LOAD_CONST MAKE_CLOSURE_0 expr GET_ITER CALL_FUNCTION_
def formatyear(self, theyear, m=3):
    for (i, row) in enumerate(self.yeardays2calendar(theyear, m)):
        names = (self.formatmonthname(theyear, k, colwidth, False)
                 for k in months)



#  Copyright (c) 2020-2022 Rocky Bernstein

from uncompyle6.scanners.tok import Token

IFELSE_STMT_RULES = frozenset(
    [
        (
            "ifelsestmt",
            (
                "testexpr",
                "c_stmts_opt",
                "jump_forward_else",
                "else_suite",
                "_come_froms",
            ),
        ),
        (
            "ifelsestmt",
            (
                "testexpr",
                "c_stmts_opt",
                "jump_forward_else",
                "else_suite",
                "\\e__come_froms",
            ),
        ),
        (
            "ifelsestmtl",
            (
                "testexpr",
                "c_stmts_opt",
                "jump_forward_else",
                "else_suitec",
            ),
        ),
        (
            "ifelsestmtc",
            (
                "testexpr",
                "c_stmts_opt",
                "jump_forward_else",
                "else_suitec",
                "\\e__come_froms",
            ),
        ),
        (
            "ifelsestmtc",
            (
                "testexpr",
                "c_stmts_opt",
                "jump_absolute_else",
                "else_suitec",
            ),
        ),
        (
            "ifelsestmtc",
            (
                "testexpr",
                "c_stmts_opt",
                "JUMP_ABSOLUTE",
                "else_suitec",
            ),
        ),
        (
            "ifelsestmt",
            (
                "testexpr",
                "c_stmts_opt",
                "jf_cfs",
                "else_suite",
                "\\e_opt_come_from_except",
            ),
        ),
        (
            "ifelsestmt",
            (
                "testexpr",
                "c_stmts_opt",
                "JUMP_FORWARD",
                "else_suite",
                "come_froms",
            ),
        ),
        (
            "ifelsestmt",
            (
                "testexpr",
                "c_stmts",
                "come_froms",
                "else_suite",
                "come_froms",
            ),
        ),
        (
            "ifelsestmt",
            (
                "testexpr",
                "c_stmts_opt",
                "jf_cfs",
                "else_suite",
                "opt_come_from_except",
            ),
        ),
        (
            "ifelsestmt",
            (
                "testexpr",
                "c_stmts_opt",
                "jf_cf_pop",
                "else_suite",
            ),
        ),
        (
            "ifelsestmt",
            (
                "testexpr",
                "stmts",
                "jf_cfs",
                "else_suite_opt",
                "opt_come_from_except",
            ),
        ),
        (
            "ifelsestmt",
            (
                "testexpr",
                "stmts",
                "jf_cfs",
                "\\e_else_suite_opt",
                "\\e_opt_come_from_except",
            ),
        ),
        (
            "ifelsestmt",
            (
                "testexpr",
                "stmts",
                "jf_cfs",
                "\\e_else_suite_opt",
                "opt_come_from_except",
            ),
        ),
    ]
)


def ifelsestmt(self, lhs, n, rule, tree, tokens, first, last):

    if (last + 1) < n and tokens[last + 1] == "COME_FROM_LOOP" and lhs != "ifelsestmtc":
        # ifelsestmt jumped outside of loop. No good.
        return True

    # print("XXX", first, last)
    # for t in range(first, last):
    #     print(tokens[t])
    # print("=" * 30)

    first_offset = tokens[first].off2int()

    if rule not in IFELSE_STMT_RULES:
        # print("XXX", rule)
        return False

    # Avoid if/else where the "then" is a "raise_stmt1" for an
    # assert statement. Parse this as an "assert" instead.
    stmts = tree[1]
    if stmts in ("c_stmts",) and len(stmts) == 1:
        raise_stmt1 = stmts[0]
        if (
                raise_stmt1 == "raise_stmt1" and
                raise_stmt1[0] in ("LOAD_ASSERT",)
        ):
            return True

    # Make sure all the offsets from the "come froms" at the
    # end of the "if" come from somewhere inside the "if".
    # Since the come_froms are ordered so that lowest
    # offset COME_FROM is last, it is sufficient to test
    # just the last one.
    if len(tree) == 5:
        end_come_froms = tree[-1]
        if end_come_froms.kind != "else_suite" and self.version >= (3, 0):
            if end_come_froms == "opt_come_from_except" and len(end_come_froms) > 0:
                end_come_froms = end_come_froms[0]
            if not isinstance(end_come_froms, Token):
                if len(end_come_froms):
                    return first_offset > end_come_froms[-1].attr
            elif first_offset > end_come_froms.attr:
                return True

        # FIXME: There is weirdness in the grammar we need to work around.
        # we need to clean up the grammar.
        if self.version < (3, 0):
            last_token = tree[-1]
        else:
            last_token = tokens[last]
        if last_token == "COME_FROM" and first_offset > last_token.attr:
            if (
                self.version < (3, 0)
                and self.insts[self.offset2inst_index[last_token.attr]].opname
                != "SETUP_LOOP"
            ):
                return True

    testexpr = tree[0]

    # Check that the condition portion of the "if"
    # jumps to the "else" part.
    if testexpr[0] in ("testtrue", "testfalse"):
        if_condition = testexpr[0]

        else_suite = tree[3]
        assert else_suite.kind.startswith("else_suite")

        if len(if_condition) > 1 and if_condition[1].kind.startswith("jmp_"):
            if last == n:
                last -= 1
            jmp = if_condition[1]
            if self.version >= (2, 7):
                jmp_target = jmp[0].attr
            else:
                jmp_target = int(jmp[0].pattr)

            # Below we check that jmp_target is jumping to a feasible
            # location. It should be to the transition after the "then"
            # block and to the beginning of the "else" block.
            # However the "if/else" is inside a loop the false test can be
            # back to the loop.

            # FIXME: the below logic for jf_cfs could probably be
            # simplified.
            jump_else_end = tree[2]
            if jump_else_end == "jf_cf_pop":
                jump_else_end = jump_else_end[0]

            if jump_else_end == "JUMP_FORWARD":
                endif_target = int(jump_else_end.pattr)
                last_offset = tokens[last].off2int()
                if endif_target != last_offset:
                    return True
            last_offset = tokens[last].off2int(prefer_last=False)
            if jmp_target == last_offset:
                # jmp_target should be jumping to the end of the if/then/else
                # but is it jumping to the beginning of the "else"
                return True
            if (
                jump_else_end in ("jf_cfs", "jump_forward_else")
                and jump_else_end[0] == "JUMP_FORWARD"
            ):
                # If the "else" jump jumps before the end of the the "if .. else end", then this
                # is not this kind of "ifelsestmt".
                jump_else_forward = jump_else_end[0]
                jump_else_forward_target = jump_else_forward.attr
                if jump_else_forward_target < last_offset:
                    return True
                pass
            if (
                jump_else_end in ("jb_elsec", "jb_elsel", "jf_cfs", "jb_cfs")
                and jump_else_end[-1] == "COME_FROM"
            ):
                if jump_else_end[-1].off2int() != jmp_target:
                    return True

            if first_offset > jmp_target:
                # A backward or loop jump from the end of an "else"
                # clause before the beginning of the "if" test is okay
                # only if we are trying to match or reduce an "if"
                # statement of the kind that can occur only inside a
                # loop construct.
                if lhs in ("ifelsestmtl", "ifelsestmtc"):
                    jump_false = jmp
                    if (
                        tree[2].kind == "JUMP_FORWARD"
                        and jump_false == "jmp_false"
                        and len(else_suite) == 1
                    ):
                        suite_stmts = else_suite[0]
                        continue_stmt = suite_stmts[0]
                        if (
                            suite_stmts == "suite_stmts"
                            and len(suite_stmts) == 1
                            and continue_stmt == "continue"
                            and jump_false[0].attr == continue_stmt[0].attr
                        ):
                            # for ...:
                            #   if ...:
                            #     ...
                            #   else:
                            #     continue
                            return False
                return True

            return (jmp_target > last_offset) and tokens[last] != "JUMP_FORWARD"

    return False



# expressions.py -- source test pattern for expressions
#
# This simple program is part of the decompyle test suite.
#
# decompyle is a Python byte-code decompiler
# See http://www.goebel-consult.de/decompyle/ for download and
# for further information

def _lsbStrToInt(str):
        return ord(str[0]) + \
               (ord(str[1]) << 8) + \
               (ord(str[2]) << 16) + \
               (ord(str[3]) << 24)



# From 3.6.4 configparser.py
# Bug in 3.6 was handling "else" with compound
# if. there is no POP_BLOCK and
# there are several COME_FROMs before the else
def _read(self, fp, a, value, f):
    for line in fp:
        for prefix in a:
            fp()
        if (value and fp and
            prefix > 5):
            f()
        else:
            f()



# From uncompyle6/verify.py
# Bug was POP_JUMP offset to short so we have a POP_JUMP
# to a JUMP_ABSOULTE and this messes up reduction rule checking.

def cmp_code_objects(member, a, tokens1, tokens2, verify, f):
    for member in members:
        while a:
            # Increase the bytecode length of the while statement
            x = 1; x = 2; x = 3; x = 4; x = 5; x = 6; x = 7; x = 8
            x = 1; x = 2; x = 3; x = 4; x = 5; x = 6; x = 7; x = 8
            x = 1; x = 2; x = 3; x = 4; x = 5; x = 6; x = 7; x = 8
            x = 1; x = 2; x = 3; x = 4; x = 5; x = 6; x = 7; x = 8
            x = 1; x = 2; x = 3; x = 4; x = 5; x = 6; x = 7; x = 8
            x = 1; x = 2; x = 3; x = 4; x = 5; x = 6; x = 7; x = 8
            x = 49; x = 50; x = 51; x = 52; x = 53;
            if tokens1:
                if tokens2:
                    continue
                elif f:
                    continue
            else:
                a = 2

            i1 += 1
            x = 54 # comment this out and we're good



# From 2.7.17 test_bdb.py
# The problem was detecting a docstring at the begining of the module
# It must be detected and change'd or else the "from __future__" below
# is invalid.
# Note that this has to be compiled with optimation < 2 or else optimization
# will remove the docstring
"""Rational, infinite-precision, real numbers."""

from __future__ import division



# Python 3.6, uses rule:
#  genexpr ::= load_closure load_genexpr LOAD_CONST
#              MAKE_FUNCTION_CLOSURE expr GET_ITER CALL_FUNCTION_1
def __sub__(self, other): # SList()-other
    return self.__class__(i for i in self if i not in other)



from uncompyle3.tests.blackbox.blackboxtestcase import BlackBoxTestCase


class TestMisc(BlackBoxTestCase):

    def test_builtin_call(self):
        infile = "misc/globals.cpython-32.pyc"
        expected = "misc/globals.py"
        self.assertProgramOutput(infile, expected)

    def test_builtin_call_dual(self):
        infile = "misc/globals_locals.cpython-32.pyc"
        expected = "misc/globals_locals.py"
        self.assertProgramOutput(infile, expected)

    def test_assign(self):
        infile = "misc/assign.cpython-32.pyc"
        expected = "misc/assign.py"
        self.assertProgramOutput(infile, expected)

    def test_assign_none(self):
        infile = "misc/assign_none.cpython-32.pyc"
        expected = "misc/assign_none.py"
        self.assertProgramOutput(infile, expected)

    def test_assign_none_str(self):
        infile = "misc/assign_none_str.cpython-32.pyc"
        expected = "misc/assign_none_str.py"
        self.assertProgramOutput(infile, expected)

    def test_complex_script1(self):
        infile = "misc/complex_script1.cpython-32.pyc"
        expected = "misc/complex_script1.py"
        self.assertProgramOutput(infile, expected)

    def test_complex_script2(self):
        infile = "misc/complex_script2.cpython-32.pyc"
        expected = "misc/complex_script2.py"
        self.assertProgramOutput(infile, expected)



# From 2.7 test_descr.py
# Testing __doc__ descriptor...
# The bug in decompilation was erroneously matching
# __doc__ = as a docstring

"""This program is self-checking!"""

def test_doc_descriptor():
    # Testing __doc__ descriptor...
    # Python SF bug 542984
    class DocDescr(object):
        def __get__(self, object, otype):
            if object:
                object = object.__class__.__name__ + ' instance'
            if otype:
                otype = otype.__name__
            return 'object=%s; type=%s' % (object, otype)
    class OldClass:
        __doc__ = DocDescr()
    class NewClass(object):
        __doc__ = DocDescr()
    assert OldClass.__doc__ == 'object=None; type=OldClass'
    assert OldClass().__doc__ == 'object=OldClass instance; type=OldClass'
    assert NewClass.__doc__ == 'object=None; type=NewClass'
    assert NewClass().__doc__ == 'object=NewClass instance; type=NewClass'

test_doc_descriptor()



if __file__:
    0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0+0


# From 3.7 test_buffer.py
# Bug is in dealing with EXTENDED_ARG instructions.
# In reduction-rule tests where we are testing the offset,
# getting *which* offset to test against, when there are two
# possible offset, can mess us up.

def five(a):
    return 5

def test_ndarray_slice_multidim(a, f, listerr):
    for slices in a:
        shape_t = 0; shape_t = 1; shape_t = 2; shape_t = 3; shape_t = 4; shape_t = 5; shape_t = 6; shape_t = 7; shape_t = 8; shape_t = 9
        shape_t = 0; shape_t = 1; shape_t = 2; shape_t = 3; shape_t = 4; shape_t = 5; shape_t = 6; shape_t = 7; shape_t = 8; shape_t = 9
        shape_t = 0; shape_t = 1; shape_t = 2; shape_t = 3; shape_t = 4; shape_t = 5; shape_t = 6; shape_t = 7; shape_t = 8; shape_t = 9
        shape_t = 0; shape_t = 1; shape_t = 2; shape_t = 3; shape_t = 4; shape_t = 5; shape_t = 6; shape_t = 7; shape_t = 8; shape_t = 9
        shape_t = 0; shape_t = 1; shape_t = 2; shape_t = 3; shape_t = 4; shape_t = 5; shape_t = 6; shape_t = 7; shape_t = 8; shape_t = 9
        nderr = None
        if nderr or listerr:
            return f(5)
        else:
            return 2

assert test_ndarray_slice_multidim([1], five, False) == 2
assert test_ndarray_slice_multidim([1], five, True) == 5

# From 3.7 test_builtin.py
def test_pow(self, m, a, b, c, f):

    shape_t = 0; shape_t = 1; shape_t = 2; shape_t = 3; shape_t = 4; shape_t = 5; shape_t = 6; shape_t = 7; shape_t = 8; shape_t = 9
    shape_t = 0; shape_t = 1; shape_t = 2; shape_t = 3; shape_t = 4; shape_t = 5; shape_t = 6; shape_t = 7; shape_t = 8; shape_t = 9
    shape_t = 0; shape_t = 1; shape_t = 2; shape_t = 3; shape_t = 4; shape_t = 5; shape_t = 6; shape_t = 7; shape_t = 8; shape_t = 9
    shape_t = 0; shape_t = 1; shape_t = 2; shape_t = 3; shape_t = 4; shape_t = 5; shape_t = 6; shape_t = 7; shape_t = 8; shape_t = 9
    shape_t = 0; shape_t = 1; shape_t = 2; shape_t = 3; shape_t = 4; shape_t = 5; shape_t = 6; shape_t = 7; shape_t = 8; shape_t = 9
    shape_t = 0; shape_t = 1; shape_t = 2; shape_t = 3; shape_t = 4; shape_t = 5; shape_t = 6; shape_t = 7; shape_t = 8; shape_t = 9

    shape_t = 0; shape_t = 1; shape_t = 2;

    for z in m:
        if a or \
           b or \
           c:
            f(TypeError)
        else:
            x = 2

    x = 3

# From 3.7 test_exceptions.py
# Bug is handling extended arg
def testAttributes(exceptionList):
    try:
        x = 0
    except:
        pass

    for exc in exceptionList:
        x = 0; x = 1; x = 2; x = 3; x = 4; x = 5; x = 6; x = 7; x = 8; x = 9;
        x = 0; x = 1; x = 2; x = 3; x = 4; x = 5; x = 6; x = 7; x = 8; x = 9;
        x = 0; x = 1; x = 2; x = 3; x = 4; x = 5; x = 6; x = 7; x = 8; x = 9;
        x = 0; x = 1; x = 2; x = 3; x = 4; x = 5; x = 6; x = 7; x = 8; x = 9;
        x = 0; x = 1; x = 2; x = 3; x = 4; x = 5; x = 6; x = 7; x = 8; x = 9;
        x = 0; x = 1; x = 2; x = 3; x = 4; x = 5; x = 6; x = 7; x = 8; x = 9;
        x = 0

# From test_urllibnet2.py
# Bug was in detecting end of "if" flowing into an extended-arg "for".
def _test_urls(retry, urls):
    if retry:
        urlopen = 1

    for url in urls:
        shape_t = 0; shape_t = 1; shape_t = 2; shape_t = 3; shape_t = 4; shape_t = 5; shape_t = 6; shape_t = 7; shape_t = 8; shape_t = 9
        shape_t = 0; shape_t = 1; shape_t = 2; shape_t = 3; shape_t = 4; shape_t = 5; shape_t = 6; shape_t = 7; shape_t = 8; shape_t = 9
        shape_t = 0; shape_t = 1; shape_t = 2; shape_t = 3; shape_t = 4; shape_t = 5; shape_t = 6; shape_t = 7; shape_t = 8; shape_t = 9
        shape_t = 0; shape_t = 1; shape_t = 2; shape_t = 3; shape_t = 4; shape_t = 5; shape_t = 6; shape_t = 7; shape_t = 8; shape_t = 9
        shape_t = 0; shape_t = 1; shape_t = 2; shape_t = 3; shape_t = 4; shape_t = 5; shape_t = 6; shape_t = 7; shape_t = 8; shape_t = 9
        shape_t = 0; shape_t = 1; shape_t = 2; shape_t = 3; shape_t = 4; shape_t = 5; shape_t = 6; shape_t = 7; shape_t = 8; shape_t = 9
        shape_t = 0



#  Copyright (c) 2019-2020, 2022 by Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Isolate Python 3.8 version-specific semantic actions here.
"""

########################
# Python 3.8+ changes
#######################

from uncompyle6.semantics.consts import PRECEDENCE, TABLE_DIRECT
from uncompyle6.semantics.customize37 import FSTRING_CONVERSION_MAP
from uncompyle6.semantics.helper import escape_string, strip_quotes

def customize_for_version38(self, version):

    # FIXME: pytest doesn't add proper keys in testing. Reinstate after we have fixed pytest.
    # for lhs in 'for forelsestmt forelselaststmt '
    #             'forelselaststmtc tryfinally38'.split():
    #     del TABLE_DIRECT[lhs]

    TABLE_DIRECT.update(
        {
            "async_for_stmt38": (
                "%|async for %c in %c:\n%+%c%-%-\n\n",
                (2, "store"),
                (0, "expr"),
                (3, ("for_block", "pass")),
            ),
            "async_forelse_stmt38": (
                "%|async for %c in %c:\n%+%c%-%|else:\n%+%c%-\n\n",
                (7, 'store'),
                (0, 'expr'),
                (8, 'for_block'),
                (-1, 'else_suite')
            ),
            "async_with_stmt38": (
                "%|async with %c:\n%+%c%-\n",
                (0, "expr"),
                (7, ("l_stmts_opt", "l_stmts", "pass")),
            ),
            "async_with_as_stmt38": (
                "%|async with %c as %c:\n%+%|%c%-",
                (0, "expr"),
                (6, "store"),
                (7, "suite_stmts"),
            ),
            "c_forelsestmt38": (
                "%|for %c in %c:\n%+%c%-%|else:\n%+%c%-\n\n",
                (2, "store"),
                (0, "expr"),
                (3, "for_block"),
                -1,
            ),
            "c_tryfinallystmt38": (
                "%|try:\n%+%c%-%|finally:\n%+%c%-\n\n",
                (1, "c_suite_stmts_opt"),
                (-2, "c_suite_stmts_opt"),
            ),
            # Python 3.8 reverses the order of keys and items
            # from all prior versions of Python.
            "dict_comp_body": ("%c: %c", (0, "expr"), (1, "expr"),),
            "except_cond1a": ("%|except %c:\n", (1, "expr"),),
            "except_cond_as": (
                "%|except %c as %c:\n",
                (1, "expr"),
                (-2, "STORE_FAST"),
            ),
            "except_handler38": ("%c", (2, "except_stmts")),
            "except_handler38a": ("%c", (-2, "stmts")),
            "except_handler38c": (
                "%c%+%c%-",
                (1, "except_cond1a"),
                (2, "except_stmts"),
            ),
            "except_handler_as": (
                "%c%+\n%+%c%-",
                (1, "except_cond_as"),
                (2, "tryfinallystmt"),
            ),
            "except_ret38a": ("return %c", (4, "expr")),
            # Note: there is a suite_stmts_opt which seems
            # to be bookkeeping which is not expressed in source code
            "except_ret38": ("%|return %c\n", (1, "expr")),
            "for38": (
                "%|for %c in %c:\n%+%c%-\n\n",
                (2, "store"),
                (0, "expr"),
                (3, "for_block"),
            ),
            "forelsestmt38": (
                "%|for %c in %c:\n%+%c%-%|else:\n%+%c%-\n\n",
                (2, "store"),
                (0, "expr"),
                (3, "for_block"),
                -1,
            ),
            "forelselaststmt38": (
                "%|for %c in %c:\n%+%c%-%|else:\n%+%c%-",
                (2, "store"),
                (0, "expr"),
                (3, "for_block"),
                -2,
            ),
            "forelselaststmtc38": (
                "%|for %c in %c:\n%+%c%-%|else:\n%+%c%-\n\n",
                (2, "store"),
                (0, "expr"),
                (3, "for_block"),
                -2,
            ),
            "ifpoplaststmtc": ("%|if %c:\n%+%c%-", (0, "testexpr"), (2, "l_stmts")),
            "pop_return": ("%|return %c\n", (1, "return_expr")),
            "popb_return": ("%|return %c\n", (0, "return_expr")),
            "pop_ex_return": ("%|return %c\n", (0, "return_expr")),
            "set_for": (" for %c in %c", (2, "store"), (0, "expr_or_arg"),),
            "whilestmt38": (
                "%|while %c:\n%+%c%-\n\n",
                (1, ("bool_op", "testexpr", "testexprc")),
                (2, ("l_stmts", "pass")),
            ),
            "whileTruestmt38": (
                "%|while True:\n%+%c%-\n\n",
                (1, ("l_stmts", "pass")),
            ),
            "try_elsestmtl38": (
                "%|try:\n%+%c%-%c%|else:\n%+%c%-",
                (1, "suite_stmts_opt"),
                (3, "except_handler38"),
                (5, "else_suitel"),
            ),
            "try_except38": (
                "%|try:\n%+%c\n%-%|except:\n%+%c%-\n\n",
                (2, ("suite_stmts_opt", "suite_stmts")),
                (3, ("except_handler38a", "except_handler38b", "except_handler38c")),
            ),
            "try_except38r": (
                "%|try:\n%+%c\n%-%|except:\n%+%c%-\n\n",
                (1, "return_except"),
                (2, "except_handler38b"),
            ),
            "try_except38r2": (
                "%|try:\n%+%c\n%-%|except:\n%+%c%c%-\n\n",
                (1, "suite_stmts_opt"),
                (8, "cond_except_stmts_opt"),
                (10, "return"),
            ),
            "try_except38r4": (
                "%|try:\n%+%c\n%-%|except:\n%+%c%c%-\n\n",
                (1, "returns_in_except"),
                (3, "except_cond1"),
                (4, "return"),
            ),
            "try_except_as": (
                "%|try:\n%+%c%-\n%|%-%c\n\n",
                (
                    -4,
                    ("suite_stmts", "_stmts"),
                ),  # Go from the end because of POP_BLOCK variation
                (-3, "except_handler_as"),
            ),
            "try_except_ret38": (
                "%|try:\n%+%c%-\n%|except:\n%+%|%c%-\n\n",
                (1, "returns"),
                (2, "except_ret38a"),
            ),
            "try_except_ret38a": (
                "%|try:\n%+%c%-%c\n\n",
                (1, "returns"),
                (2, "except_handler38c"),
            ),
            "tryfinally38rstmt": (
                "%|try:\n%+%c%-%|finally:\n%+%c%-\n\n",
                (0, "sf_pb_call_returns"),
                (-1, ("ss_end_finally", "suite_stmts", "_stmts")),
            ),
            "tryfinally38rstmt2": (
                "%|try:\n%+%c%-%|finally:\n%+%c%-\n\n",
                (4, "returns"),
                -2,
                "ss_end_finally",
            ),
            "tryfinally38rstmt3": (
                "%|try:\n%+%|return %c%-\n%|finally:\n%+%c%-\n\n",
                (1, "expr"),
                (-1, "ss_end_finally"),
            ),
            "tryfinally38rstmt4": (
                "%|try:\n%+%c%-\n%|finally:\n%+%c%-\n\n",
                (1, "suite_stmts_opt"),
                (5, "suite_stmts_return"),
            ),
            "tryfinally38stmt": (
                "%|try:\n%+%c%-%|finally:\n%+%c%-\n\n",
                (1, "suite_stmts_opt"),
                (6, "suite_stmts_opt"),
            ),
            "tryfinally38astmt": (
                "%|try:\n%+%c%-%|finally:\n%+%c%-\n\n",
                (2, "suite_stmts_opt"),
                (8, "suite_stmts_opt"),
            ),
            "named_expr": (  # AKA "walrus operator"
                "%c := %p",
                (2, "store"),
                (0, "expr", PRECEDENCE["named_expr"] - 1),
            ),
        }
    )

    def except_return_value(node):
        if node[0] == "POP_BLOCK":
            self.default(node[1])
        else:
            self.template_engine(("%|return %c\n", (0, "expr")), node)
        self.prune()

    self.n_except_return_value = except_return_value

    # FIXME: now that we've split out cond_except_stmt,
    # we should be able to get this working as a pure transformation rule,
    # so no procedure is needed here.
    def try_except38r3(node):
        self.template_engine(("%|try:\n%+%c\n%-", (1, "suite_stmts_opt")), node)
        cond_except_stmts_opt = node[5]
        assert cond_except_stmts_opt == "cond_except_stmts_opt"
        for child in cond_except_stmts_opt:
            if child == "cond_except_stmt":
                if child[0] == "except_cond1":
                    self.template_engine(
                        ("%c\n", (0, "except_cond1"), (1, "expr")), child
                    )
                    self.template_engine(("%+%c%-\n", (1, "except_stmts")), child)
                pass
            pass
        self.template_engine(("%+%c%-\n", (7, "return")), node)
        self.prune()

    self.n_try_except38r3 = try_except38r3

    def n_list_afor(node):
        if len(node) == 2:
            # list_afor ::= get_iter list_afor
            self.comprehension_walk_newer(node, 0)
        else:
            list_iter_index = 2 if node[2] == "list_iter" else 3
            self.template_engine(
                (
                    " async for %[1]{%c} in %c%[1]{%c}",
                    (1, "store"),
                    (0, "get_aiter"),
                    (list_iter_index, "list_iter"),
                ),
                node,
            )
        self.prune()

    self.n_list_afor = n_list_afor

    def n_set_afor(node):
        if len(node) == 2:
            self.template_engine(
                (" async for %[1]{%c} in %c", (1, "store"), (0, "get_aiter")), node
            )
        else:
            self.template_engine(
                " async for %[1]{%c} in %c%c",
                (1, "store"),
                (0, "get_aiter"),
                (2, "set_iter"),
            )
        self.prune()

    self.n_set_afor = n_set_afor

    def n_formatted_value_debug(node):
        p = self.prec
        self.prec = 100

        formatted_value = node[1]
        value_equal = node[0].attr
        assert formatted_value.kind.startswith("formatted_value")
        old_in_format_string = self.in_format_string
        self.in_format_string = formatted_value.kind
        format_value_attr = node[-1]

        post_str = ""
        if node[-1] == "BUILD_STRING_3":
            post_load_str = node[-2]
            post_str = self.traverse(post_load_str, indent="")
            post_str = strip_quotes(post_str)

        if format_value_attr == "FORMAT_VALUE_ATTR":
            attr = format_value_attr.attr
            if attr & 4:
                fmt = strip_quotes(self.traverse(node[3], indent=""))
                attr_flags = attr & 3
                if attr_flags:
                    conversion = "%s:%s" % (
                        FSTRING_CONVERSION_MAP.get(attr_flags, ""),
                        fmt,
                    )
                else:
                    conversion = ":%s" % fmt
            else:
                conversion = FSTRING_CONVERSION_MAP.get(attr, "")
            f_str = "f%s" % escape_string(
                "{%s%s}%s" % (value_equal, conversion, post_str)
            )
        else:
            f_conversion = self.traverse(formatted_value, indent="")
            # Remove leaving "f" and quotes
            conversion = strip_quotes(f_conversion[1:])
            f_str = "f%s" % escape_string(f"{value_equal}{conversion}" + post_str)

        self.write(f_str)
        self.in_format_string = old_in_format_string

        self.prec = p
        self.prune()

    self.n_formatted_value_debug = n_formatted_value_debug

    def n_suite_stmts_return(node):
        if len(node) > 1:
            assert len(node) == 2
            self.template_engine(
                ("%c\n%|return %c", (0, ("_stmts", "suite_stmts")), (1, "expr")), node
            )
        else:
            self.template_engine(("%|return %c", (0, "expr")), node)
        self.prune()

    self.n_suite_stmts_return = n_suite_stmts_return



from uncompyle3.tests.blackbox.blackboxtestcase import BlackBoxTestCase


class TestLooping(BlackBoxTestCase):

    def test_while(self):
        infile = "looping/while.cpython-32.pyc"
        expected = "looping/while.py"
        self.assertProgramOutput(infile, expected)

    def test_for(self):
        infile = "looping/for.cpython-32.pyc"
        expected = "looping/for.py"
        self.assertProgramOutput(infile, expected)



#  Copyright (c) 2022 Rocky Bernstein
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


def whilestmt38_check(
    self, lhs: str, n: int, rule, ast, tokens: list, first: int, last: int
) -> bool:
    # When we are missing a COME_FROM_LOOP, the
    # "while" statement is nested inside an if/else
    # so after the POP_BLOCK we have a JUMP_FORWARD which forms the "else" portion of the "if"
    # Check this.
    # print("XXX", first, last, rule)
    # for t in range(first, last):
    #     print(tokens[t])
    # print("=" * 40)

    if tokens[last] != "COME_FROM" and tokens[last - 1] == "COME_FROM":
        last -= 1
    if tokens[last - 1].kind.startswith("RAISE_VARARGS"):
        return True
    while tokens[last] == "COME_FROM":
        last -= 1
    # In a "while" loop, (in contrast to "for" loop), the loop jump is
    # always to the first offset
    first_offset = tokens[first].off2int()
    if tokens[last] == "JUMP_LOOP" and (
        tokens[last].attr == first_offset or tokens[last - 1].attr == first_offset
    ):
        return False
    return True



#  Copyright (c) 2017, 2021-2022 by Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Python 3.5 bytecode decompiler scanner

Does some additional massaging of xdis-disassembled instructions to
make things easier for decompilation.

This sets up opcodes Python's 3.5 and calls a generalized
scanner routine for Python 3.
"""

from __future__ import print_function

from uncompyle6.scanners.scanner3 import Scanner3

# bytecode verification, verify(), uses JUMP_OPs from here
from xdis.opcodes import opcode_35 as opc
JUMP_OPS = opc.JUMP_OPS

class Scanner35(Scanner3):

    def __init__(self, show_asm=None, is_pypy=False):
        Scanner3.__init__(self, (3, 5), show_asm, is_pypy)
        return
    pass

if __name__ == "__main__":
    from xdis.version_info import PYTHON_VERSION_TRIPLE, version_tuple_to_str

    if PYTHON_VERSION_TRIPLE[:2] == (3, 5):
        import inspect

        co = inspect.currentframe().f_code  # type: ignore
        tokens, customize = Scanner35().ingest(co)
        for t in tokens:
            print(t.format())
        pass
    else:
        print("Need to be Python 3.5 to demo; I am version %s" % version_tuple_to_str())



# Bug from python 3.x handling set comprehensions
{y for y in range(3)}

# Bug in python 3.4 (base64.py) in handling dict comprehension
b = {v: k for k, v in enumerate(b3)}

# Bug from Python 3.4 enum
def __new__(classdict):
    members = {k: classdict[k] for k in classdict._member_names}
    return members

# Bug from 3.5.1 enum.py in 2.7, and 3.x
{a for b in bases for a in b.__dict__}



# Tests:

# return_expr_or_cond ::= return_expr
# if_exp_ret ::= expr POP_JUMP_IF_FALSE expr RETURN_END_IF return_expr_or_cond
# return_expr_or_cond ::= if_exp_ret
# ret_or ::= expr JUMP_IF_TRUE_OR_POP return_expr_or_cond COME_FROM

# See https://github.com/rocky/python-uncompyle6/issues/5

def minimize(x, y):
    return x or (x if x < y else y)



# From 2.6 test_datetime.py
# Bug is in parsing (x is 0 or x is 1) and (y is 5 or y is 2)
# correctly.

# This code is RUNNABLE!
result = []
for y in (1, 2, 10):
    x = cmp(1, y)
    if (x is 0 or x is 1) and (y is 5 or y is 2):
        expected = 10
    elif y is 2:
        expected = 2
    else:
        expected = 3
    result.append(expected)

assert result == [3, 2, 3]



import pytest
# uncompyle6
from xdis.version_info import PYTHON_VERSION_TRIPLE, IS_PYPY
from validate import validate_uncompyle


@pytest.mark.skipif(PYTHON_VERSION_TRIPLE < (3, 6) or IS_PYPY, reason="need at least Python 3.6 and not PyPY")
@pytest.mark.parametrize('text', (
    "{0.: 'a', -1: 'b'}",   # BUILD_MAP
    "{'a':'b'}",            # BUILD_MAP
    "{0: 1}",               # BUILD_MAP
    "{b'0':1, b'2':3}",     # BUILD_CONST_KEY_MAP
    "{0: 1, 2: 3}",         # BUILD_CONST_KEY_MAP
    "{'a':'b','c':'d'}",    # BUILD_CONST_KEY_MAP
    "{0: 1, 2: 3}",         # BUILD_CONST_KEY_MAP
    "{'a': 1, 'b': 2}",     # BUILD_CONST_KEY_MAP
    "{'a':'b','c':'d'}",    # BUILD_CONST_KEY_MAP
    "{0.0:'b',0.1:'d'}",    # BUILD_CONST_KEY_MAP
))
def test_build_const_key_map(text):
    validate_uncompyle(text)



from uncompyle3.tests.blackbox.blackboxtestcase import BlackBoxTestCase


class TestOpUnary(BlackBoxTestCase):

    def test_positive(self):
        infile = "operation_unary/positive.cpython-32.pyc"
        expected = "operation_unary/positive.py"
        self.assertProgramOutput(infile, expected)

    def test_negative(self):
        infile = "operation_unary/negative.cpython-32.pyc"
        expected = "operation_unary/negative.py"
        self.assertProgramOutput(infile, expected)

    def test_not(self):
        infile = "operation_unary/not.cpython-32.pyc"
        expected = "operation_unary/not.py"
        self.assertProgramOutput(infile, expected)

    def test_invert(self):
        infile = "operation_unary/invert.cpython-32.pyc"
        expected = "operation_unary/invert.py"
        self.assertProgramOutput(infile, expected)

    def test_iter(self):
        infile = "operation_unary/iter.cpython-32.pyc"
        expected = "operation_unary/iter.py"
        self.assertProgramOutput(infile, expected)



# Self-checking test.
# Mixed boolean expresions

b = True
assert b, 'b = True'
c = False
assert not c, 'c = False'
d = True
a = b and c or d
assert a, 'b and c or d'
a = (b or c) and d
assert a, '(b or c) and d'
a = b or c or d
assert  a, 'b or c or d'
a = b and c and d
assert not a, 'b and c and d'
a = b or c and d
assert a
a = b and (c or d)
assert a
a = b and c or d
assert a
a = (b or c and d) and b
assert a
a = (b or c and d or a) and b
assert a



#  Copyright (c) 2022 Rocky Bernstein

from uncompyle6.scanners.tok import Token


def for_block_invalid(self, lhs, n, rule, tree, tokens, first: int, last: int) -> bool:

    # print("XXX", first, last)
    # for t in range(first, last):
    #     print(tokens[t])
    # print("=" * 30)

    if rule == (
        "for_block",
        ("l_stmts_opt", "JUMP_ABSOLUTE", "JUMP_BACK", "JUMP_BACK"),
    ):
        # Check that the two JUMP_BACK's go to the same place.
        jump_back1 = tokens[last - 2]
        jump_back2 = tokens[last - 1]
        if jump_back1.attr != jump_back2.attr:
            return True

        # Also check that JUMP_ABSOLUTE jumps to the JUMP_BACK.
        # In this situation the JUMP_ABSOLUTE and a JUMP_BACK
        # is not needed, but it seems to be there anyway.

        jump_absolute = tokens[last - 3]
        if jump_absolute.attr != jump_back2.offset:
            return True

        # Right now all of this is known to happen only in Python 2.7.
        if self.version[:2] == (2, 7):
            return False

    if len(rule[1]) <= 1 or not tree:
        return False

    come_froms = tree[-1]
    # This is complicated, but note that the JUMP_IF instruction comes immediately
    # *before* _ifstmts_jump so that's what we have to test
    # the COME_FROM against. This can be complicated by intervening
    # POP_TOP, and pseudo COME_FROM, ELSE instructions
    #
    pop_jump_index = first - 1
    while pop_jump_index > 0 and tokens[pop_jump_index] in (
        "ELSE",
        "POP_TOP",
        "JUMP_FORWARD",
        "COME_FROM",
    ):
        pop_jump_index -= 1

    # FIXME: something is fishy when and EXTENDED ARG is needed before the
    # pop_jump_index instruction to get the argment. In this case, the
    # _ifsmtst_jump can jump to a spot beyond the come_froms.
    # That is going on in the non-EXTENDED_ARG case is that the POP_JUMP_IF
    # jumps to a JUMP_(FORWARD) which is changed into an EXTENDED_ARG POP_JUMP_IF
    # to the jumped forwarded address
    if tokens[pop_jump_index].attr > 256:
        return False

    pop_jump_offset = tokens[pop_jump_index].off2int(prefer_last=False)
    if isinstance(come_froms, Token):
        if tokens[pop_jump_index].attr < pop_jump_offset and tree[0] != "pass":
            # This is a jump backwards to a loop. All bets are off here when there the
            # unless statement is "pass" which has no instructions associated with it.
            return False
        return come_froms.attr is not None and pop_jump_offset > come_froms.attr

    elif len(come_froms) == 0:
        return False
    else:
        return pop_jump_offset > come_froms[-1].attr



#  Copyright (c) 2022 by Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Isolate Python 1.4- version-specific semantic actions here.
"""

from uncompyle6.semantics.consts import TABLE_DIRECT

#######################
# Python 1.4- Changes #
#######################
def customize_for_version14(self, version):
    TABLE_DIRECT.update(
        {
            "print_expr_stmt": (
                ("%|print %c\n", 0)
            ),
        }
    )



# 2.6.9 asynchat.py
# 2.6 added:
# return_stmt ::= return_expr RETURN_END_IF come_from_pop
def initiate_send(self, x):
    while self:
        if self:
            x = 'a'
        else:
            del self.producer_fifo[0]
        return x

# 2.6.9 contextlib.py
# Bug was in return exc, so added:
# return_stmt ::= return_expr RETURN_VALUE_IF come_from_pop

def __exit__(self, type, value, traceback):
    try:
        raise RuntimeError
    except StopIteration:
        return value
    except:
        raise



#  Copyright (c) 2015-2022 by Rocky Bernstein
#  Copyright (c) 2000-2002 by hartmut Goebel <h.goebel@crazy-compilers.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
All the crazy things we have to do to handle Python functions in Python before 3.0.
The saga of changes continues in 3.0 and above and in other files.
"""
from typing import List, Tuple
from uncompyle6.scanner import Code
from uncompyle6.semantics.parser_error import ParserError
from uncompyle6.parser import ParserError as ParserError2
from uncompyle6.semantics.helper import (
    print_docstring,
    find_all_globals,
    find_globals_and_nonlocals,
    find_none,
)
from xdis import iscode

def make_function1(self, node, is_lambda, nested=1, code_node=None):
    """
    Dump function defintion, doc string, and function body.
    This code is specialied for Python 2.
    """

    def build_param(tree, param_names: List[str]) -> Tuple[bool, List[str]]:
        """build parameters:
            - handle defaults
            - handle format tuple parameters
        """
        # if formal parameter is a tuple, the paramater name
        # starts with a dot (eg. '.1', '.2')
        args = tree[0]
        del tree[0]
        params = []
        assert args.kind in ("star_args", "args", "varargs")
        has_star_arg = args.kind in ("star_args", "varargs")
        args_store = args[2]
        if args_store == "args_store":
            for arg in args_store:
                params.append(param_names[arg.attr])
        return has_star_arg, params

    # MAKE_FUNCTION_... or MAKE_CLOSURE_...
    assert node[-1].kind.startswith("BUILD_")

    defparams = []
    # args_node = node[-1]
    # if isinstance(args_node.attr, tuple):
    #     # positional args are after kwargs
    #     defparams = node[1 : args_node.attr[0] + 1]
    #     pos_args, kw_args, annotate_argc = args_node.attr
    # else:
    #     defparams = node[: args_node.attr]
    #     kw_args = 0
    #     pass

    lambda_index = None

    if lambda_index and is_lambda and iscode(node[lambda_index].attr):
        assert node[lambda_index].kind == "LOAD_LAMBDA"
        code = node[lambda_index].attr
    else:
        code = code_node.attr

    assert iscode(code)
    code = Code(code, self.scanner, self.currentclass)

    # add defaults values to parameter names
    argc = code.co_argcount
    paramnames = list(code.co_varnames[:argc])

    # defaults are for last n parameters, thus reverse
    paramnames.reverse()
    defparams.reverse()

    try:
        tree = self.build_ast(
            code._tokens,
            code._customize,
            code,
            is_lambda=is_lambda,
            noneInNames=("None" in code.co_names),
        )
    except (ParserError, ParserError2) as p:
        self.write(str(p))
        if not self.tolerate_errors:
            self.ERROR = p
        return

    indent = self.indent

    # build parameters
    has_star_arg, params = build_param(tree, code.co_names)

    if has_star_arg:
        params[-1] = "*" + params[-1]

    # dump parameter list (with default values)
    if is_lambda:
        self.write("lambda ", ", ".join(params))
        # If the last statement is None (which is the
        # same thing as "return None" in a lambda) and the
        # next to last statement is a "yield". Then we want to
        # drop the (return) None since that was just put there
        # to have something to after the yield finishes.
        # FIXME: this is a bit hoaky and not general
        if (
            len(tree) > 1
            and self.traverse(tree[-1]) == "None"
            and self.traverse(tree[-2]).strip().startswith("yield")
        ):
            del tree[-1]
            # Now pick out the expr part of the last statement
            tree_expr = tree[-1]
            while tree_expr.kind != "expr":
                tree_expr = tree_expr[0]
            tree[-1] = tree_expr
            pass
    else:
        self.write("(", ", ".join(params))

    # if kw_args > 0:
    #     if not (4 & code.co_flags):
    #         if argc > 0:
    #             self.write(", *, ")
    #         else:
    #             self.write("*, ")
    #         pass
    #     else:
    #         self.write(", ")

    #     for n in node:
    #         if n == "pos_arg":
    #             continue
    #         else:
    #             self.preorder(n)
    #         break
    #     pass

    # if code_has_star_star_arg(code):
    #     if argc > 0:
    #         self.write(", ")
    #     self.write("**%s" % code.co_varnames[argc + kw_pairs])

    if is_lambda:
        self.write(": ")
    else:
        self.println("):")

    if (
        len(code.co_consts) > 0 and code.co_consts[0] is not None and not is_lambda
    ):  # ugly
        # docstring exists, dump it
        print_docstring(self, indent, code.co_consts[0])

    if not is_lambda:
        assert tree == "stmts"

    all_globals = find_all_globals(tree, set())

    globals, nonlocals = find_globals_and_nonlocals(
        tree, set(), set(), code, self.version
    )

    # Python 1 doesn't support the "nonlocal" statement

    for g in sorted((all_globals & self.mod_globs) | globals):
        self.println(self.indent, "global ", g)
    self.mod_globs -= all_globals
    has_none = "None" in code.co_names
    rn = has_none and not find_none(tree)
    tree.code = code
    self.gen_source(
        tree, code.co_name, code._customize, is_lambda=is_lambda, returnNone=rn
    )

    code._tokens = None # save memory
    code._customize = None  # save memory



import pytest
from uncompyle6 import code_deparse
from xdis.version_info import PYTHON_VERSION_TRIPLE

pytest.mark.skip(PYTHON_VERSION_TRIPLE < (2, 7), reason="need Python < 2.7")


def test_single_mode():
    single_expressions = (
        "i = 1",
        "i and (j or k)",
        "i += 1",
        "i = j % 4",
        "i = {}",
        "i = []",
        "for i in range(10):\n    i\n",
        "for i in range(10):\n    for j in range(10):\n        i + j\n",
        # 'try:\n    i\nexcept Exception:\n    j\nelse:\n    k\n'
    )

    for expr in single_expressions:
        code = compile(expr + "\n", "<string>", "single")
        got = code_deparse(code, compile_mode="single").text
        assert got == expr + "\n"



#  Copyright (c) 2016-2017, 2022 Rocky Bernstein
"""
spark grammar differences over Python 3.2 for Python 3.1.
"""
from __future__ import print_function

from uncompyle6.parser import PythonParserSingle
from uncompyle6.parsers.parse32 import Python32Parser


class Python31Parser(Python32Parser):
    def p_31(self, args):
        """
        subscript2     ::= expr expr DUP_TOPX BINARY_SUBSCR

        setupwith      ::= DUP_TOP LOAD_ATTR store LOAD_ATTR CALL_FUNCTION_0 POP_TOP
        setupwithas    ::= DUP_TOP LOAD_ATTR store LOAD_ATTR CALL_FUNCTION_0 store
        with           ::= expr setupwith SETUP_FINALLY
                           suite_stmts_opt
                           POP_BLOCK LOAD_CONST COME_FROM_FINALLY
                           load delete WITH_CLEANUP END_FINALLY

        # Keeps Python 3.1 "with .. as" designator in the same position as it is in other version.
        setupwithas31  ::= setupwithas SETUP_FINALLY load delete

        withasstmt     ::= expr setupwithas31 store
                           suite_stmts_opt
                           POP_BLOCK LOAD_CONST COME_FROM_FINALLY
                           load delete WITH_CLEANUP END_FINALLY

        store ::= STORE_NAME
        load  ::= LOAD_FAST
        load  ::= LOAD_NAME
        """

    def remove_rules_31(self):
        self.remove_rules(
            """
        # DUP_TOP_TWO is DUP_TOPX in 3.1 and earlier
        subscript2 ::= expr expr DUP_TOP_TWO BINARY_SUBSCR

        # The were found using grammar coverage
        list_if     ::= expr jmp_false list_iter COME_FROM
        list_if_not ::= expr jmp_true list_iter COME_FROM
        """
        )

    def customize_grammar_rules(self, tokens, customize):
        super(Python31Parser, self).customize_grammar_rules(tokens, customize)
        self.remove_rules_31()
        return

    pass


class Python31ParserSingle(Python31Parser, PythonParserSingle):
    pass


if __name__ == "__main__":
    # Check grammar
    p = Python31Parser()
    p.remove_rules_31()
    p.check_grammar()
    from xdis.version_info import IS_PYPY, PYTHON_VERSION_TRIPLE

    if PYTHON_VERSION_TRIPLE[:2] == (3, 1):
        lhs, rhs, tokens, right_recursive, dup_rhs = p.check_sets()
        from uncompyle6.scanner import get_scanner

        s = get_scanner(PYTHON_VERSION_TRIPLE, IS_PYPY)
        opcode_set = set(s.opc.opname).union(
            set(
                """JUMP_BACK CONTINUE RETURN_END_IF COME_FROM
               LOAD_GENEXPR LOAD_ASSERT LOAD_SETCOMP LOAD_DICTCOMP LOAD_CLASSNAME
               LAMBDA_MARKER RETURN_LAST
            """.split()
            )
        )
        # FIXME: try this
        remain_tokens = set(tokens) - opcode_set
        import re

        remain_tokens = set([re.sub(r"_\d+$", "", t) for t in remain_tokens])
        remain_tokens = set([re.sub("_CONT$", "", t) for t in remain_tokens])
        remain_tokens = set(remain_tokens) - opcode_set
        print(remain_tokens)
        import sys

        if len(sys.argv) > 1:
            from spark_parser.spark import rule2str

            for rule in sorted(p.rule2name.items()):
                print(rule2str(rule[0]))



# Github Issue #57 with Python 2.7
def some_function():
    return ['some_string']

def some_other_function():
    some_variable, = some_function()
    print(some_variable)

# From 2.7 test_compile.py
# Bug is adding erroneous parens in d[(1:2, 1:2)] += 1
def bug(d):
    d[1:2, 1:2] += 1

empty_tup = ()
one_item_tup = ("item1", )
one_item_tup_without_parentheses = "item",
many_items_tup = ("item1", "item2", "item3")



# From uncompyle issue #295 on 3.8
# In 3.8 BREAK_LOOP and CONTINUE_LOOP no longer exist.
# The bug is that the "break" is turned into:
# POP_BLOCK
# JUMP_ABSOLUTE
while True:
    try:
        x = 1
        break
    except Exception:
        pass

while True:
    try:
        x -= 1
    except Exception:
        break

# Issue #25 https://github.com/rocky/python-decompile3/issues/25
# Same as above using "for".
for i in range(5):
    try:
        x = 1
        break
    except Exception:
        if i == 4:
            raise

# From 3.8.1 _osx_support.py
# Bug was handling a "break" inside a "try".
# In 3.8 POP_EXCEPT is moved before "JUMP_ABSOLUTE" of
# the break.
# FIXME:
# def compiler_fixup(compiler_so, cc_args, index):
#     if index:
#         while True:
#             try:
#                 index = 1
#             except:
#                 index = 2
#                 break


# Bug was returning an IfExp inside "with":
# the return value is mixed in with the "with"
# code finalization.
# FIXME:
# def _read_output(x, a):
#     with x as fp:
#         return fp if a else None


# In 3.8 the CONTINUE_LOOP disappears,
# and this makes it harder to detect continue
# inside a loop with a continue in the except clause.
def connect_ws_with_retry(f1, f2):
    while True:
        try:
            f1()
        except Exception:
            f2()
            continue



#!/usr/bin/env python
"""byte compiles a Python 1.x program"""
import sys
if len(sys.argv) != 2:
    print("Usage: compile-file.py *python-file*")
    sys.exit(1)
source = sys.argv[1]

# assert source.endswith('.py')
basename = source[:-3]

# We do this crazy way to support Python 1.4 which
# doesn't support version_info.
PY_VERSION = sys.version[:3]

bytecode = "%s-%s.pyc" % (basename, PY_VERSION)

import py_compile
print("# compiling %s to %s" % (source, bytecode))
py_compile.compile(source, bytecode)
# import os
# os.system("../bin/uncompyle6 %s" % bytecode)



"""This program is self-checking!"""
# Python 2.4 - 2.7 bug in transforming "else if" to "elif" in Python 2.4 .. 2.7
# From Issue #377

# RUNNABLE!
def loop_continue_dead_code(slots):
    for name in slots:
        if name:
            pass
        else:
            continue
            # The below is dead code
            if x:
                y()
            else:
                z()

loop_continue_dead_code([None, 1])
